/************************************************************************
 * Programa    : mceitor 0.9.2 			         		*
 * Archivo     : mxlib.c  						*
 * Descripcion : Funciones de creacion, operacion y manejo de variables *
 * tipo MATRIX.								*
 * 									*
 * Autor : Boris Estudiez						*
 * Nick  : slice							*
 * Fecha : 09/11/2001 							*
 * Mails : 43824@electronica.frc.utn.edu.ar - boris@flashmail.com 	*
 * Requiere: Libreria matematica (libm)					*
 *	 								*
 * Licenced Under GNU							*
 *	 								*
 *	 								*
 * Powered by C!							*
 *  									*
 ************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include "include/mxlib.h" /* Localizacion de mxlib.h */

extern double MIN_NUMP; /* se define en mceitor.c */

/*------------------FUNCIONES DE MANEJO DE MATRICES---------------------*/

/************************************************************************
 *  new_matrix: crea una variable tipo MATRIX en memoria.		*
 *  Devuelve NULL si no se puede crear o la direccion de la variable	*
 *  tipo MATRIX creada.							*
 *  Asigna dimenciones a mat->row y mat->col.				*
 *  Se debe usar esta funcion para crear matrices siempre.		*
 *  row y col deben ser > 0.						*
 ************************************************************************/

MATRIX *new_matrix(int row, int col)
{         
   
   MATRIX *mat;
   int i;
   
   if(row < 0 || col < 0 || row == 0 || col == 0) 
      return NULL; /* filas y columnas incorrectas */
    
   if((mat = (MATRIX *) malloc(sizeof(MATRIX))) == NULL)
      return NULL;
  
   /*Creamos matriz en memoria*/
   if((mat->val = (double **) calloc(row, sizeof(double))) == NULL)
      return NULL;
   for(i=0; i < row; i++)
      if((mat->val[i] = (double *) calloc(col, sizeof(double))) == NULL)
         return NULL;
   
   /*Guardamos en la variable tipo MATRIX las dimenciones de la matriz*/
   mat->row = row; 
   mat->col = col;
   
   return mat;
}

/************************************************************************
 * del_matrix: Libera la memoria ocupada por una variable tipo MATRIX   *
 ************************************************************************/

void del_matrix(MATRIX *mat)
{
   int i;
   
   for(i=0; i < mat->row; i++)
      free(mat->val[i]);
   free(mat->val);
   free(mat);   
}   

/************************************************************************
 * colcpy_matrix: Copia de mat2 la columna col2 a mat1 en la col. col1.	*
 * Ambas tienen que tener igual numero de filas.			*
 * Devuelve M_ERROR si hay error o M_OK de lo contrario.		*
 ************************************************************************/
 
int colcpy_matrix(MATRIX *mat1, MATRIX *mat2, int col1, int col2)
{
   int i;
   
   if(mat1->row != mat2->row)
      return M_ERROR;  /* distinta cantidad de filas */

   if(col1 >= mat1->col || col2 >= mat2->col || col1 < 0 || col2 < 0) 
      return M_ERROR; /*fuera de rango*/   
   
   for(i=0; i < mat2->row; i++)
      mat1->val[i][col1] = mat2->val[i][col2];
   return M_OK;   
}   

/************************************************************************
 * cpy_matrix: copia mat2 a mat1. Devuelve M_ERROR en caso de tener	*
 * dimenciones distintas o M_OK.					*
 ************************************************************************/
int cpy_matrix(MATRIX *mat1, MATRIX *mat2)
{

   int i, j; /* i fila, j columna */
   
   if(mat1->row != mat2->row || mat1->col != mat2->col)
      return M_ERROR; /*dimenciones distintas*/
   
   for(i=0; i < mat2->row; i++)
      for(j=0; j < mat2->col; j++)
         mat1->val[i][j] = mat2->val[i][j];
         
}

/*-----------------------FUNCIONES MATEMATICAS---------------------------*/

/************************************************************************
 * trimat: Triangula una matriz nxn a traves de operaciones elementales.*
 * Retorna el producto de todos los escalares usados en las operaciones *
 * elementales de e3 por el producto de tantos -1 como operaciones	*
 * elementales e1 se le hagan a la matriz para trinagularla.            *
 * Retorna 1 en caso de no efectuar operaciones e1 y e3.		*
 * Retorna M_ERROR en caso de error.					* 
 ************************************************************************/

double trimat(MATRIX *mat)
{
   double k, value=1;   
   int i, j; /*numero de fila y col. actual */
   int i2; /* variable que albergara un numero de fila */
   
   if(mat->row != mat->col)
      return M_ERROR; /*matriz no cuadrada*/
     
   for(i = 0, j = 0; i < mat->row && j < mat->col; i++, j++){
      if(mat->val[i][j] == 0){ /* primer elemento nulo */
         if((i2 = scandval(mat, i, j)) == M_NOFOUND)
            continue; /* todos los elementos por debajo son nulos */
         else {
            e1(mat, i, i2); /* f2 no tiene elemento nulo en esta columna*/  
            value *= -1;
         }   
      }
      if(mat->val[i][j] != 1){ 
         k = mat->val[i][j];
         e3(mat, i, 1/k); /* convertimos en 1 el elemento de esta col. */           
         value *= k;
      }
      if(i < (mat->row -1)) /* no estamos en la ultima fila*/
         while((i2 = scandval(mat, i+1, j)) != M_NOFOUND){ 
            k = mat->val[i2][j] * -1;
            e2(mat, i2, i, k);
         }   
   }
   return value;
}

/*************************************************************************
 * detm: calcula el determinate de una matriz por el metodo de triangular*
 * la matriz, luego el determinate asociado a tal matriz es el prod-	 *
 * ucto de la diagonal principal por las propiedades que tienen		 *
 * las operaciones elementales sobre los determinantes.			 *
 * Retorna el valor del determinante.					 *
 * El retorno para el error no esta definido.				 *
 *	 								 *
 *************************************************************************/
double detm(MATRIX *mat)
{
   MATRIX *m_temp; /*copia de la matriz original*/
   double detval;
   int i;
   
   if(mat->col != mat->row) /* Matriz no cuadrada, error */
      return M_ERROR;
   
   if((m_temp = new_matrix(mat->row, mat->col)) == NULL)
      return M_ERROR;
      
   cpy_matrix(m_temp, mat);  /* copiamos matriz orignal a temporal */    
   detval = trimat(m_temp); /* triangulamos matriz y obtenemos prop. 
      		    	     de operaciones elementales */   

   for(i = 0; i < mat->row; i++)
      detval *= m_temp->val[i][i]; /*producto de la diagonal principal*/

   del_matrix(m_temp); /*liberamos matriz temporal*/   
   return (double) detval;
}   


/************************************************************************
 * e1: intercambia i1 por i2 						*
 ************************************************************************/
 
void e1(MATRIX *mat, int i1, int i2)
{
   double temp;
   int j;
   
   for(j = 0; j < mat->col; j++){
      temp = mat->val[i1][j];
      mat->val[i1][j] = mat->val[i2][j];
      mat->val[i2][j] = temp;
   }
}

/************************************************************************
 * e2: suma a i1 la i2 previamente multiplicada por k 			*
 ************************************************************************/

void e2(MATRIX *mat, int i1, int i2, double k)
{
   int j;
   
   for(j = 0; j < mat->col; j++){
      mat->val[i1][j] += mat->val[i2][j] * k;
      if(fabs(mat->val[i1][j]) < MIN_NUMP) /* redondear a 0 */
         mat->val[i1][j] = 0;      
   }      

}

/************************************************************************
 * e3: multiplica i por k 						*
 ************************************************************************/
 
void e3(MATRIX *mat, int i, double k)
{
   int j;
   
   for(j = 0; j < mat->col; j++){
      mat->val[i][j] *= k;
      if(fabs(mat->val[i][j]) < MIN_NUMP) /* redondear a 0 */
         mat->val[i][j] = 0;
   }      
}      

/************************************************************************ 
 * scandval: busca el primer valor distinto de cero desde i1 hasta 	*
 * mat->row traves de la columna col de la matriz. Devuelve M_NOFOUND en*
 * caso de no encontrar ningun valor distinto a cero. 			*
 * Si encuentra un valor != 0 regresa la fila donde esta el valor. 	*
 ************************************************************************/
   
int scandval(MATRIX *mat, int i1, int col)
{
   int i2; /* fila donde estara el elemento no nulo */
   
   for(i2 = i1; i2 < mat->row; i2++)
      if(mat->val[i2][col] != 0)
         return i2;
   return M_NOFOUND;      
}         


/************************************************************************
 * scanuval: busca el primer valor distinto de cero desde i1 hasta 0 a	*
 * traves de la columna col de la matriz. Devuelve M_NOFOUND en caso de *
 * no encontrar ningun valor o la fila donde se encuentra el valor.	*
 ************************************************************************/
   
int scanuval(MATRIX *mat, int i1, int col)
{
   int i2;
   
   for(i2 = i1; i2 > -1; i2--)
      if(mat->val[i2][col] != 0)
         return i2;
   return M_NOFOUND;      
}         


/************************************************************************
 * scanrval: busca el primer valor distinto de cero desde c1 hasta 	*
 * mat->col a traves de la fila 'row' de la matriz. Devuelve M_NOFOUND *
 * en caso de no encontrar ningun valor o la fila donde se encuentra el	*
 * valor.								*
 ************************************************************************/
   
int scanrval(MATRIX *mat, int row, int c1)
{
   int c2;
   
   for(c2 = c1; c2 < mat->col; c2++)
      if(mat->val[row][c2] != 0)
         return c2;
   return M_NOFOUND;      
}         

/************************************************************************
 * nullups: Busca en la matriz una fila que tiene mas cantidad de ceros	*
 * precedentes a algun elemento no nulo que la fila 'crow' (current row)*
 * y que esta por encima de la misma. 					*
 * ej:									*
 * 00001 (fila 1)							*
 * 00100 (fila 2) 							*
 * 10000 (fila 3)							*
 * Llamar a la funcion desde la fila 2, retornara el numero de fila = 1 *
 * 									*
 * Regresa el numero de fila o M_NOFOUND en caso de no encontrar.	*
 ************************************************************************/
 
int nullups(MATRIX *mat, int crow)
{
   int i, j; /* i fila, j columna */
   int nz=0, nzcrow=0; /*numero de ceros actuales y numero de ceros de crow*/
   
   
   for(j=0; j < mat->col; j++){ /*ceros de la fila actual */
      if(mat->val[crow][j] != 0)
         break;
      ++nzcrow;
   }
         
   for(i=0; i < crow; i++){
      for(j=0, nz=0; j < mat->col; j++){
         if(mat->val[i][j] != 0)
            break;
         ++nz;   
      }   
      if(nz > nzcrow)
         return i; /*fila con mayor numero de ceros arriba de la acual*/
   }
   
   /*no hay fila con mayor numeros de ceros precedentes arriba*/
   return M_NOFOUND;
}      

/************************************************************************
 * sysresolv: resuelve un sistema de n ecuaciones y n incognitas.	*
 * Almacena los valores en m_result, devuelve M_NOSOL, si el sistema	*
 * no tiene solucion unica, M_ERROR para error y M_OK de lo contrario.	*
 * El metodo a utilizar es el de Cramer. xn = detm(xn)/detm(A) .		*
 ************************************************************************/
 
int sysresolv(MATRIX *mA, MATRIX *mb, MATRIX *m_result)  
{
   MATRIX *m_temp; /* matriz temporal */
   double detA, detsus;
   int i, j; /* i fila, j columna */
   
   
   /*La matriz mb y m_result deben tener la misma catidad de filas que mA*/
   if(mA->row != mb->row || mA->row != m_result->row)
      return M_ERROR;
   
   /* primero evaluamos si el sistema tiene solucion unica */
   if((detA = detm(mA)) == 0)
      return M_NOSOL; 
         
   /* matriz temporal */   
   if((m_temp = new_matrix(mA->row, 1)) == NULL)
      return M_ERROR;
 
   
   /*Aplicamos el metodo de Cramer*/
   
   for(j = 0; j < mA->row; j++){
      i = j;
      colcpy_matrix(m_temp, mA, 0, j); /* copia en m_temp de la columna j de mA */
      colcpy_matrix(mA, mb, j, 0);
      detsus = detm(mA);
      m_result->val[i][0] = (detsus / detA); /*valor de las incognitas*/
      colcpy_matrix(mA, m_temp, j, 0); /*volvemos a mA con sus valores originales*/   
   } 
   
   del_matrix(m_temp); /*liberamos matriz temporal*/   
   return M_OK;
}   
   

/************************************************************************
 * mcomp: determina el menor complementario del elemento en [row,col].	*
 * La matriz debe ser cuadrada. Para un error su retorno no esta	*
 * definido.								*
 ************************************************************************/

double mcomp(MATRIX *mat, int row, int col)
{
   int i, j; /* indices para recorrer la matriz mat*/
   int tc=0, tr=0; /* indices para recorrer la matriz m_temp*/
   double mc; /*menor complementario*/
   MATRIX *m_temp; /*matriz temporal*/
   
   if(mat->col != mat->row)
      return M_ERROR; /* matriz no cuadrada */
   
   if((m_temp = new_matrix(mat->row-1, mat->col-1)) == NULL)
      return M_ERROR;

   /*copiamos el menor complementario de mat a m_temp */
   for(i=0; i < mat->row; i++)
      for(j=0; j < mat->col; j++){
         if(j == col || i == row)
            continue;
         if(tc == m_temp->col){
            tc = 0;
            ++tr;
         }   
         m_temp->val[tr][tc++] = mat->val[i][j];
      }     
         
   mc = detm(m_temp);
   del_matrix(m_temp); /*liberamos espacio temporal*/
   
   return mc;      
}  

/************************************************************************
 * cof: convierte a mat en su matriz cofactor. Devuelve M_ERR si 	*
 * hay error y M_OK si no lo hay.					*
 ************************************************************************/
  
int cof(MATRIX *mat)
{       
        
   MATRIX *m_temp;
   int i, j;
      
   if(mat->row != mat->col)
      return M_ERROR; /*matriz no cuadrada*/

   if((m_temp = new_matrix(mat->row, mat->col)) == NULL)
      return M_ERROR;

   cpy_matrix(m_temp, mat); /*copiamos mat a m_temp*/
   
   for(i=0; i < mat->row; i++)
      for(j=0; j < mat->col; j++)
         mat->val[i][j] =  pow(-1, i+1+j+1) * mcomp(m_temp, i, j); 
   
   del_matrix(m_temp); /*liberamos espacio temporal*/
   return M_OK;
}         

/************************************************************************
 * trn: Transpone una matriz mat. 					*
 * Si la matriz no es cuadrada, la redimenciona.			*
 * M_OK = ok - M_ERR = error						*
 ************************************************************************/   

int trn(MATRIX *mat){

   MATRIX *m_temp;
   int i, j;

   if((m_temp = new_matrix(mat->row, mat->col)) == NULL)
      return M_ERROR;

   cpy_matrix(m_temp, mat); /*copiamos mat a m_temp*/
      
   if(mat->row != mat->col){ /*matriz no cuadrada, la redimencionamos*/
      del_matrix(mat);
      if((mat = new_matrix(mat->col, mat->row)) == NULL)
         return M_ERROR;
   }   
  
   for(i=0; i < m_temp->row; i++)
      for(j=0; j < m_temp->col; j++)
         mat->val[j][i] = m_temp->val[i][j];
   
   del_matrix(m_temp);      
}


/************************************************************************
 *adj: convierte a mat en su matriz adjunta, la cual por definicion es 	*
 *la transpuesta de la matriz cofactor. M_ERROR Al retornar error.		*
 ************************************************************************/

int adj(MATRIX *mat)
{

   if(cof(mat) == M_ERROR)
      return M_ERROR;
   trn(mat);
   
   return M_OK; 
}      

/************************************************************************
 * multk: Multiplica una matriz por un escalar k			*
 ************************************************************************/
 
void multk(MATRIX *mat, double k)
{
   int i, j;
   
   for(i=0; i < mat->row; i++)
      for(j=0; j < mat->col; j++)
         mat->val[i][j] *= k;
}


/************************************************************************      
 * Calcula la inversa de una matriz por el metodo de la adjunta.	*
 * inv(A) = adj(A)/detm(A). Devuelve M_NOSOL al no poder invertirla 	*
 * porque el detm(A) = 0 (inversible) M_ERROR, en caso de error.	*
 ************************************************************************/

int inv(MATRIX *mat)
{

   double detval;
   
   if(mat->row != mat->col)
      return M_ERROR;
   
   if((detval = detm(mat)) == 0)
      return M_NOSOL; /*matriz inversible*/
      
   adj(mat); /*hacemos su adjunta*/
   multk(mat, 1/detval); /* metodo de la adjunta*/
   
   return M_OK; 
}   
      

/************************************************************************   
 * redmat: Reduce la matriz mat a su reducida por filas.		*
 * Esta funcion se encuentra en un estado BETA.				*
 * El arreglo de punteros op, contendra todas la op elementales echas   *
 * a la matriz, tal arreglo debe ser suficientemente grande para conte- *
 * ner a todas las operaciones elementales. 				*
 * Si no se pasa segundo parametro, este debe ser NULL.			*
 ************************************************************************/

void redmat(MATRIX *mat, char **op)
{
   
   double k;
   int i, j; /*numero de fila y col. actual */
   int i2;
   int n=0; /* numero de operaciones elementales */
   char buf[BUFSIZ];
        
   for(i=0, j=0; j < mat->col && i < mat->row; i++, j++){ 
      if(mat->val[i][j] == 0){ /* primer elemento nulo de fila */
         if(scanrval(mat, i, j+1) == M_NOFOUND && 
            scandval(mat, i+1, j) == M_NOFOUND)
            continue; 	/*fila nula, y por debajo en col. j solo ceros*/
         while(j < mat->col){ 	/*buscamos un elemento por debajo*/
             /*hay un elemento abajo?*/
            if((i2 = scandval(mat, i, j)) != M_NOFOUND){
               e1(mat, i, i2); /* intercambiamos i2 por i*/  
               if(op != NULL){
                  sprintf(buf, "e1(%d, %d)", i+1, i2+1);
                  op[n] = (char *) calloc(strlen(buf)+1, sizeof(char));
                  strcpy(op[n], buf);
                  n++; 
               }   
               break;
            }
            ++j;
         }      
      }
      if(mat->val[i][j] != 1){ /*convertimos en 1 el elemento de esta col.*/      
         k = mat->val[i][j];
         e3(mat, i, 1/k);      
         if(op != NULL){
            sprintf(buf, "e3(%d, %g)", i+1, 1/k);
            op[n] = (char *) calloc(strlen(buf)+1, sizeof(char));
            strcpy(op[n], buf);
            n++; 
        }    
      }
      if(i < (mat->row -1)) /* no estamos en la ultima fila*/
         while((i2 = scandval(mat, i+1, j)) != M_NOFOUND){ 
            k = mat->val[i2][j] * -1;
            e2(mat, i2, i, k);
            if(op != NULL){
               sprintf(buf, "e2(%d, %d, %g)", i2+1, i+1, k);
               op[n] = (char *) calloc(strlen(buf)+1, sizeof(char));                        
               strcpy(op[n], buf);
               n++; 
            }   
         }
      if(i > 0) /* no estamos en la primera fila*/
         while((i2 = scanuval(mat, i-1, j)) != M_NOFOUND){ 
            k = mat->val[i2][j] * -1;
            e2(mat, i2, i, k);
            if(op != NULL){
               sprintf(buf, "e2(%d, %d, %g)", i2+1, i+1, k);
               op[n] = (char *) calloc(strlen(buf)+1, sizeof(char));            
               strcpy(op[n], buf);
               n++;
            }     
         }   
      if((i2 = nullups(mat, i)) != M_NOFOUND){ /*la fila de arriba tiene mas ceros*/
         e1(mat, i, i2); /*intercambiamos las filas*/
         if(op != NULL){
            sprintf(buf, "e1(%d, %d)", i+1, i2+1);
            op[n] = (char *) calloc(strlen(buf)+1, sizeof(char));         
            strcpy(op[n], buf);
            n++;
         }    
         i = i2;
      }   
   }
   op[n]=NULL; /* arreglo terminado en null */
}
   






